#include <stdio.h>
#include <stdint.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <getopt.h>
#include <stddef.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <linux/types.h>
#include <linux/spi/spidev.h>

#define FRAME_MAX_SIZE      8

/**
 * mode: 0 -- CPOL=0,CPHA=0
 *       1 -- CPOL=0,CPHA=1
 *       2 -- CPOL=1,CPHA=0
 *       3 -- CPOL=1,CPHA=1
 */
static uint8_t  mode  = 0;

/**
 * bits: 1 ~ 16
 */
static uint8_t  bits  = 8;

/**
 * speed: <= 72MHz
 */
static uint32_t speed = 1 * 1000 * 1000;

/**
 * device: 0/1
 */
static const char *device = "/dev/spidev0.0";

static uint16_t delay = 0;

static int spi_fd = -1;

/**
 * function: open mspi device
 */
int ss_mspi_open(void)
{
    int ret = 0;
    uint8_t lsb;

    if (spi_fd >= 0)
        return -1;

    spi_fd = open(device, O_RDWR);
    if (spi_fd < 0)
    {
        printf("can't open %s\n", device);
        return -1;
    }

    //set mspi mode
    ret = ioctl(spi_fd, SPI_IOC_WR_MODE, &mode);
    if (ret < 0)
    {
        printf("can't set spi mode\n");
        close(spi_fd);
        return -1;
    }

    ret = ioctl(spi_fd, SPI_IOC_RD_MODE, &mode);
    if (ret < 0)
    {
        printf("can't get spi mode\n");
        close(spi_fd);
        return -1;
    }

    //set bits per word
    ret = ioctl(spi_fd, SPI_IOC_WR_BITS_PER_WORD, &bits);
    if (ret < 0)
    {
        printf("can't set bits per word\n");
        close(spi_fd);
        return -1;
    }

    ret = ioctl(spi_fd, SPI_IOC_RD_BITS_PER_WORD, &bits);
    if (ret < 0)
    {
        printf("can't get bits per word\n");
        close(spi_fd);
        return -1;
    }

    //set mspi speed hz
    ret = ioctl(spi_fd, SPI_IOC_WR_MAX_SPEED_HZ, &speed);
    if (ret < 0)
    {
        printf("can't set max speed hz\n");
        close(spi_fd);
        return -1;
    }

    ret = ioctl(spi_fd, SPI_IOC_RD_MAX_SPEED_HZ, &speed);
    if (ret < 0)
    {
        printf("can't get max speed hz\n");
        close(spi_fd);
        return -1;
    }

    //ret = ioctl(spi_fd, SPI_IOC_WR_LSB_FIRST, &lsb);
    //if (ret < 0)
    //{
    //    printf("mspi set lsb first falid!\n");
    //    close(spi_fd);
    //    return -1;
    //}

    ret = ioctl(spi_fd, SPI_IOC_RD_LSB_FIRST, &lsb);
    if (ret < 0)
    {
        printf("mspi get lsb first falid!\n");
        close(spi_fd);
        return -1;
    }

    printf("mspi mode: %d\n", mode);
    printf("mspi bits per word: %d\n", bits);
    printf("mspi speed: %d Hz\n", speed);
    printf("mspi transtmit is lsb first: %d\n", lsb);

    return 0;
}


/**
 * function: close mspi device
 */
int ss_mspi_close(void)
{
    if (spi_fd < 0)
        return -1;

    close(spi_fd);
    spi_fd = -1;

    printf("close %s success\n", device);

    return 0;
}

/**
 * function: mspi send data
 */
int ss_mspi_write(uint8_t *wr_buf, int size)
{
    int i, j, ret;
    uint16_t tx[FRAME_MAX_SIZE] = {0};  //max 8
    int frame_cnt = size / sizeof(tx), data_cnt = size % sizeof(tx);

    if (spi_fd < 0)
        return -1;

    struct spi_ioc_transfer tr = {
        .tx_buf = (unsigned long)tx,
        .rx_buf = NULL,
        .len = sizeof(tx),
        .delay_usecs = delay,
        .speed_hz = speed,
        .bits_per_word = bits,
    };

    for (i = 0; i < frame_cnt; i ++)
    {
        tr.len = sizeof(tx);
        memcpy((uint8_t *)tx, (uint8_t *)wr_buf + i * sizeof(tx), sizeof(tx));
        ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &tr);
        if (ret < 0)
        {
            printf("can't send spi message");
            return -1;
        }
    }

    if (data_cnt)
    {
        tr.len = data_cnt;
        memset((uint8_t *)tx, 0x0, sizeof(tx));
        memcpy((uint8_t *)tx, (uint8_t *)wr_buf + size - data_cnt, data_cnt);
        ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &tr);
        if (ret < 0)
        {
            printf("can't send spi message");
            return -1;
        }
    }

    return 0;
}


/**
 * function: mspi read data
 */
int ss_mspi_read(uint8_t *rd_buf, int size)
{
    int i, j, ret;
    uint16_t rx[FRAME_MAX_SIZE] = {0};
    int frame_cnt = size / sizeof(rx), data_cnt = size % sizeof(rx);

    if (spi_fd < 0)
        return -1;

    struct spi_ioc_transfer tr = {
        .tx_buf = NULL,
        .rx_buf = (unsigned long)rx,
        .len = sizeof(rx),
        .delay_usecs = delay,
        .speed_hz = speed,
        .bits_per_word = bits,
    };

    for (i = 0; i < frame_cnt; i ++)
    {
        tr.len = sizeof(rx);
        ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &tr);
        if (ret < 0)
        {
            printf("can't send spi message");
            return -1;
        }
        memcpy((uint8_t *)rd_buf + i * sizeof(rx), (uint8_t *)rx, sizeof(rx));
    }

    if (data_cnt)
    {
        tr.len = data_cnt;
        memset((uint8_t *)rx, 0x0, sizeof(rx));
        ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &tr);
        if (ret < 0)
        {
            printf("can't send spi message");
            return -1;
        }
        memcpy((uint8_t *)rd_buf + size - data_cnt, (uint8_t *)rx, data_cnt);
    }

    return 0;
}

/**
 * function: mspi lookback test
 */
int ss_mspi_lookback(uint8_t *wr_buf, uint8_t *rd_buf, int size)
{
    int i, j, ret;
    uint16_t rx[FRAME_MAX_SIZE] = {0};
    uint16_t tx[FRAME_MAX_SIZE] = {0};
    int frame_cnt = size / sizeof(tx), data_cnt = size % sizeof(tx);
    //printf("frame count: %d, data count: %d\n", frame_cnt, data_cnt);

    if (spi_fd < 0)
        return -1;

    struct spi_ioc_transfer tr = {
        .tx_buf = (unsigned long)tx,
        .rx_buf = (unsigned long)rx,
        .len = sizeof(tx),
        .delay_usecs = delay,
        .speed_hz = speed,
        .bits_per_word = bits,
    };

    for (i = 0; i < frame_cnt; i ++)
    {
        tr.len = sizeof(tx);
        memcpy((uint8_t *)tx, (uint8_t *)wr_buf + i * sizeof(tx), sizeof(tx));
        //printf("mspi tx_buf: %04x %04x %04x %04x %04x %04x %04x %04x\n", tx[0], tx[1], tx[2], tx[3], tx[4], tx[5], tx[6], tx[7]);

        ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &tr);
        if (ret < 0)
        {
            printf("can't send spi message");
            return -1;
        }

        memcpy((uint8_t *)rd_buf + i * sizeof(rx), (uint8_t *)rx, sizeof(rx));
        //printf("mspi rx_buf: %04x %04x %04x %04x %04x %04x %04x %04x\n", rx[0], rx[1], rx[2], rx[3], rx[4], rx[5], rx[6], rx[7]);
    }

    if (data_cnt)
    {
        tr.len = data_cnt;
        memset((uint8_t *)tx, 0x0, sizeof(tx));
        memset((uint8_t *)rx, 0x0, sizeof(rx));
        memcpy((uint8_t *)tx, (uint8_t *)wr_buf + size - data_cnt, data_cnt);
        //printf("mspi tx_buf: %04x %04x %04x %04x %04x %04x %04x %04x\n", tx[0], tx[1], tx[2], tx[3], tx[4], tx[5], tx[6], tx[7]);

        ret = ioctl(spi_fd, SPI_IOC_MESSAGE(1), &tr);
        if (ret < 0)
        {
            printf("can't send spi message");
            return -1;
        }

        memcpy((uint8_t *)rd_buf + size - data_cnt, (uint8_t *)rx, data_cnt);
        //printf("mspi rx_buf: %04x %04x %04x %04x %04x %04x %04x %04x\n", rx[0], rx[1], rx[2], rx[3], rx[4], rx[5], rx[6], rx[7]);
    }

    return 0;
}

int main(int argc, char *argv[])
{
    int i, ret, len;
    uint8_t tmp_buf[64];
    uint8_t wr_buf[64], rd_buf[64];

    if (argc < 3)
    {
        printf("please inpu such as: ./mspi_main bits string\n");
        printf("eg: ./mspi_main 8 FF\n");
        printf("eg: ./mspi_main 9 01FF\n");
        return -1;
    }

    printf("mspi test start!\n");

    //set mspi transmit bits
    bits = atoi(argv[1]);

    ret = ss_mspi_open();
    if (ret < 0)
        return -1;

    memset(wr_buf, 0x0, sizeof(wr_buf));
    memset(rd_buf, 0x0, sizeof(rd_buf));
    memset(tmp_buf, 0x0, sizeof(rd_buf));

    //receive shell '0'~'F' string and convert to char data
    uint8_t *cp = argv[2];
    uint8_t tmp = 0;
    for (i = 0; *cp; i ++, cp ++)
    {
        tmp = *cp - '0';
        if (tmp > 9)
            tmp -= ('A' - '0') - 10;
        if (tmp > 15)
            tmp -= ('a' - 'A');
        if (tmp > 15)
        {
            printf("invalid hex data on %c!\n", *cp);
            ss_mspi_close();
            return -1;
        }

        if (i % 2)
            tmp_buf[i / 2] |= tmp;
        else
            tmp_buf[i / 2]  = tmp << 4;
    }
    len = i / 2;
    printf("test len: %d, bits: %d, in data: %s\n", len, bits, argv[2]);

    //save data to an array
    if (bits <= 8)
    {
        //save data to uint8_t
        for (i = 0; i < len; i ++)
        {
            wr_buf[i] = tmp_buf[i];
        }
    }
    else
    {
        //need to convert to uint16_t, high bits save to high address
        for (i = 0; i < len; i += 2)
        {
            wr_buf[i + 1] = tmp_buf[i];
            wr_buf[i] = tmp_buf[i + 1];
        }
    }

    for (i = 0; i < len; i ++)
    {
        if (!(i % 8))
            printf("\ninput  data: ");
        printf("0x%02x ", wr_buf[i]);
    }
    printf("\n\n");

    ret = ss_mspi_lookback(wr_buf, rd_buf, len);
    if (ret < 0)
    {
        ss_mspi_close();
        return -1;
    }

    for (i = 0; i < len; i ++)
    {
        if (!(i % 8))
            printf("\noutput data: ");
        printf("0x%02x ", rd_buf[i]);
    }
    printf("\n");

    ss_mspi_close();

    return 0;
}

